#!/usr/bin/env python3

import pytest
import autoprop
from functools import partial

def _test_policy(make_policy_test_cls, class_decorator, getter_decorator, perturb, recalculate, inherit_perturbers):
    MyObj = make_policy_test_cls(
            class_decorator,
            getter_decorator,
            inherit_perturbers,
    )

    o1 = MyObj(1)
    o2 = MyObj(2)

    assert o1.x == o1.get_x() == 1
    assert o2.x == o2.get_x() == 2

    # Indirectly modify the return value of `MyObj.get_x()` via a mutable 
    # dictionary.  This avoids triggering any of the conditions that would 
    # invalidate the cache.
    o1.d['x'] = 3

    perturb(o1)

    assert o1.x == o1.get_x() == (3 if recalculate else 1)
    assert o2.x == o2.get_x() == 2

def make_policy_test_cls(class_decorator, getter_decorator, inherit_perturbers, include_setters=True, include_refresh=True):

    if inherit_perturbers:
        @autoprop.dynamic
        class MyParent:

            if include_setters:
                def set_x(self, x):
                    pass

                def del_x(self):
                    pass

                def set_y(self, y):
                    pass

                def del_y(self):
                    pass

            if include_refresh:
                @autoprop.refresh
                def refresh(self):
                    pass

                @classmethod
                @autoprop.refresh
                def refresh_cls(cls):
                    pass

                @staticmethod
                @autoprop.refresh
                def refresh_static():
                    pass

            def method(self):
                pass

            @classmethod
            def method_cls(cls):
                pass

            @staticmethod
            def method_static():
                pass

        @class_decorator
        class MyObj(MyParent):

            def __init__(self, x):
                self.d = {'x': x}

            @getter_decorator
            def get_x(self):
                return self.d['x']

    else:
        @class_decorator
        class MyObj:

            def __init__(self, x):
                self.d = {'x': x}

            @getter_decorator
            def get_x(self):
                return self.d['x']

            if include_setters:
                def set_x(self, x):
                    pass

                def del_x(self):
                    pass

                def set_y(self, y):
                    pass

                def del_y(self):
                    pass

            if include_refresh:
                @autoprop.refresh
                def refresh(self):
                    pass

                @classmethod
                @autoprop.refresh
                def refresh_cls(cls):
                    pass

                @staticmethod
                @autoprop.refresh
                def refresh_static():
                    pass

            def method(self):
                pass

            @classmethod
            def method_cls(cls):
                pass

            @staticmethod
            def method_static():
                pass

    return MyObj

def make_policy_decorators(policy):
    return [
            (
                autoprop.cache(policy=policy),
                noop,
            ), (
                autoprop.cache,
                autoprop.cache(policy=policy),
            ), (
                autoprop.cache(),
                autoprop.cache(policy=policy),
            ), (
                autoprop.cache,
                autoprop.policy(policy),
            )
    ]

def set_x(obj):
    obj.x = 0

def del_x(obj):
    del obj.x

def set_y(obj):
    obj.y = 0

def del_y(obj):
    del obj.y

def set_z(obj):
    obj.z = 0

def refresh(obj):
    obj.refresh()

def refresh_cls(obj):
    obj.refresh_cls()

def refresh_static(obj):
    obj.refresh_static()

def method(obj):
    obj.method()

def method_cls(obj):
    obj.method_cls()

def method_static(obj):
    obj.method_static()

def noop(obj):
    return obj


@pytest.mark.parametrize(
        'class_decorator, getter_decorator', [
            *make_policy_decorators('immutable'),
            (autoprop.dynamic, autoprop.immutable),
            (autoprop.immutable, noop),
        ],
)
@pytest.mark.parametrize(
        'perturb, recalculate', [
            (autoprop.clear_cache, True),
            (set_z, False),
            (refresh, False),
            (refresh_cls, False),
            (refresh_static, False),
            (method, False),
            (method_cls, False),
            (method_static, False),
            (noop, False),
        ],
)
@pytest.mark.parametrize(
        'inherit_perturbers', [False, True],
)
def test_policy_immutable(class_decorator, getter_decorator, perturb, recalculate, inherit_perturbers):
    _test_policy(
            partial(make_policy_test_cls, include_setters=False),
            class_decorator,
            getter_decorator,
            perturb,
            recalculate,
            inherit_perturbers,
    )

@pytest.mark.parametrize(
        'class_decorator, getter_decorator', [
            *make_policy_decorators('property'),
            (autoprop.cache, noop),
            (autoprop.cache(), noop),
        ],
)
@pytest.mark.parametrize(
        'perturb, recalculate', [
            (autoprop.clear_cache, True),
            (set_x, True),
            (del_x, True),
            (set_y, False),
            (del_y, False),
            (set_z, False),
            (refresh, False),
            (refresh_cls, False),
            (refresh_static, False),
            (method, False),
            (method_cls, False),
            (method_static, False),
            (noop, False),
        ],
)
@pytest.mark.parametrize(
        'inherit_perturbers', [False, True],
)
def test_policy_property(class_decorator, getter_decorator, perturb, recalculate, inherit_perturbers):
    _test_policy(
            make_policy_test_cls,
            class_decorator,
            getter_decorator,
            perturb,
            recalculate,
            inherit_perturbers,
    )

@pytest.mark.parametrize(
        'class_decorator, getter_decorator',
            make_policy_decorators('object'),
)
@pytest.mark.parametrize(
        'perturb, recalculate', [
            (autoprop.clear_cache, True),
            (set_x, True),
            (del_x, True),
            (set_y, True),
            (del_y, True),
            (set_z, True),
            (refresh, True),
            (refresh_cls, False),
            (refresh_static, False),
            (method, False),
            (method_cls, False),
            (method_static, False),
            (noop, False),
        ],
)
@pytest.mark.parametrize(
        'inherit_perturbers', [False, True],
)
def test_policy_object(class_decorator, getter_decorator, perturb, recalculate, inherit_perturbers):
    _test_policy(
            make_policy_test_cls,
            class_decorator,
            getter_decorator,
            perturb,
            recalculate,
            inherit_perturbers,
    )

@pytest.mark.parametrize(
        'class_decorator, getter_decorator', 
            make_policy_decorators('class'),
)
@pytest.mark.parametrize(
        'perturb, recalculate', [
            (autoprop.clear_cache, True),
            (set_x, True),
            (del_x, True),
            (set_y, True),
            (del_y, True),
            (set_z, True),
            (refresh, True),
            (refresh_cls, True),
            (refresh_static, True),
            (method, False),
            (method_cls, False),
            (method_static, False),
            (noop, False),
        ],
)
@pytest.mark.parametrize(
        'inherit_perturbers', [False, True],
)
def test_policy_class(class_decorator, getter_decorator, perturb, recalculate, inherit_perturbers):
    _test_policy(
            make_policy_test_cls,
            class_decorator,
            getter_decorator,
            perturb,
            recalculate,
            inherit_perturbers,
    )

@pytest.mark.parametrize(
        'class_decorator, getter_decorator', [
            *make_policy_decorators('dynamic'),
            (autoprop.cache, autoprop.dynamic),
            (autoprop.dynamic, noop),
        ],
)
@pytest.mark.parametrize(
        'perturb, recalculate', [
            (autoprop.clear_cache, True),
            (set_x, True),
            (del_x, True),
            (set_y, True),
            (del_y, True),
            (set_z, True),
            (refresh, True),
            (refresh_cls, True),
            (refresh_static, True),
            (method, True),
            (method_cls, True),
            (method_static, True),
            (noop, True),
        ],
)
@pytest.mark.parametrize(
        'inherit_perturbers', [False, True],
)
def test_policy_dynamic(class_decorator, getter_decorator, perturb, recalculate, inherit_perturbers):
    _test_policy(
            make_policy_test_cls,
            class_decorator,
            getter_decorator,
            perturb,
            recalculate,
            inherit_perturbers,
    )

@pytest.mark.parametrize(
        'class_decorator, getter_decorator', [
            (autoprop, noop)
        ],
)
@pytest.mark.parametrize(
        'perturb, recalculate', [
            (autoprop.clear_cache, True),
            (set_x, True),
            (del_x, True),
            (set_y, True),
            (del_y, True),
            (set_z, True),
            (method, True),
            (method_cls, True),
            (method_static, True),
            (noop, True),
        ],
)
@pytest.mark.parametrize(
        'inherit_perturbers', [False, True],
)
def test_policy_no_cache(class_decorator, getter_decorator, perturb, recalculate, inherit_perturbers):
    _test_policy(
            partial(make_policy_test_cls, include_refresh=False),
            class_decorator,
            getter_decorator,
            perturb,
            recalculate,
            inherit_perturbers,
    )


def test_cache_inheritance():
    # The child class determines whether or not the attribute is cached.

    @autoprop.dynamic
    class Parent:

        def __init__(self, x): #
            self.d = {'x': x}

        @autoprop.cache
        def get_super_cache(self): #
            return self.d['x']

        @autoprop.cache
        def get_parent_cache(self): #
            return self.d['x']

        def get_child_cache(self): #
            return self.d['x']

        @autoprop.cache
        def get_both_cache(self): #
            return self.d['x']

    @autoprop.dynamic
    class Child(Parent):

        def get_parent_cache(self): #
            return self.d['x']

        @autoprop.cache
        def get_child_cache(self): #
            return self.d['x']

        @autoprop.cache
        def get_both_cache(self): #
            return self.d['x']

    p = Parent(1)
    c = Child(2)

    assert p.super_cache == 1
    assert p.parent_cache == 1
    assert p.child_cache == 1
    assert p.both_cache == 1

    assert c.super_cache == 2
    assert c.parent_cache == 2
    assert c.child_cache == 2
    assert c.both_cache == 2

    p.d['x'] = 3
    c.d['x'] = 4

    assert p.super_cache == 1
    assert p.parent_cache == 1
    assert p.child_cache == 3
    assert p.both_cache == 1

    assert c.super_cache == 2
    assert c.parent_cache == 4
    assert c.child_cache == 2
    assert c.both_cache == 2

def test_inherit_uncached():
    # Note that `Child` defines `set_x()` but not `get_x()`.  This means that 
    # it will create an `x` property, but it will use `Parent.get_x()` when 
    # doing so.  Because `get_x()` is defined in a class without caching, that 
    # behavior should extend to the new property, even though getters defined 
    # in `Child` are cached by default.

    @autoprop
    class Parent:

        def __init__(self, x):
            self.d = {'x': x}

        def get_x(self):
            return self.d['x']


    @autoprop.cache
    class Child(Parent):

        def set_x(self, x):
            pass

        def get_y(self):
            return self.d['x']


    p = Parent(1)
    c = Child(2)

    assert p.x == 1
    assert c.x == 2
    assert c.y == 2

    p.d['x'] = 3
    c.d['x'] = 4

    assert p.x == 3  # not cached
    assert c.x == 4  # not cached
    assert c.y == 2  # cached

def test_dont_cache_getters_with_extra_args():
    # Normally the getter function and the property are both cached, but 
    # caching has to be disabled in the case where the getter is directly 
    # called with optional arguments.

    @autoprop.cache
    class MyObj:

        def get_x(self, i=0):
            return self.d['x'][i]

    obj = MyObj()
    obj.d = {'x': [1, 2]}

    assert obj.x == 1
    assert obj.get_x() == 1
    assert obj.get_x(0) == 1
    assert obj.get_x(1) == 2

    obj.d['x'] = [3, 4]

    assert obj.x == 1
    assert obj.get_x() == 1
    assert obj.get_x(0) == 3
    assert obj.get_x(1) == 4

def test_cache_disabled_getter_err():
    with pytest.raises(ValueError) as err:

        @autoprop
        class MyObj:

            @autoprop.cache
            def get_x(self):
                pass

    assert err.match(r"cache disabled: can't specify cache policies")
    assert err.match(r"class: .*MyObj")
    assert err.match(r"method: .*MyObj.get_x")

def test_cache_disabled_refresh_err():
    with pytest.raises(ValueError) as err:

        @autoprop
        class MyObj:

            @autoprop.refresh
            def method(self):
                pass

    assert err.match(r"cache disabled: can't use `@autoprop.refresh`")
    assert err.match(r"class: .*MyObj")
    assert err.match(r"method: .*MyObj.method")

def test_cache_disabled_refresh_cls_err():
    with pytest.raises(ValueError) as err:

        @autoprop
        class MyObj:

            @classmethod
            @autoprop.refresh
            def method(cls):
                pass

    assert err.match(r"cache disabled: can't use `@autoprop.refresh`")
    assert err.match(r"class: .*MyObj")
    assert err.match(r"method: .*MyObj.method")

def test_read_only_setter_err():
    with pytest.raises(ValueError) as err:

        @autoprop.cache
        class MyObj:

            @autoprop.cache(policy='immutable')
            def get_x(self):
                pass

            def set_x(self, x):
                pass

    assert err.match(r"can't specify setter for immutable property")
    assert err.match(r"property: .*MyObj\.x")
    assert err.match(r"immutable getter: .*MyObj\.get_x")
    assert err.match(r"setter: .*MyObj.set_x")

def test_read_only_deleter_err():
    with pytest.raises(ValueError) as err:

        @autoprop.cache
        class MyObj:

            @autoprop.cache(policy='immutable')
            def get_x(self):
                pass

            def del_x(self):
                pass

    assert err.match(r"can't specify deleter for immutable property")
    assert err.match(r"property: .*MyObj\.x")
    assert err.match(r"immutable getter: .*MyObj\.get_x")
    assert err.match(r"deleter: .*MyObj.del_x")

def test_cache_non_getter_err():
    with pytest.raises(ValueError) as err:

        @autoprop.cache
        class MyObj:

            @autoprop.cache
            def non_getter(self):
                pass

    assert err.match(r"can't cache .*MyObj\.non_getter; it's not a getter")

def test_unknown_policy_err():
    with pytest.raises(ValueError) as err:

        @autoprop.cache(policy='unknown')
        class MyObj:
            pass

    assert err.match(r"unknown policy 'unknown', expected one of")

    with pytest.raises(ValueError) as err:

        @autoprop.cache
        class MyObj:

            @autoprop.cache(policy='unknown')
            def get_x(self):
                pass

    assert err.match(r"unknown policy 'unknown', expected one of")




#!/usr/bin/env python

import pytest
import autoprop

def test_metadata():
    assert hasattr(autoprop, '__doc__')
    assert hasattr(autoprop, '__version__')

def test_get():
    @autoprop
    class Example: #
        def get_attr(self): #
            return 'attr'

    ex = Example()
    assert ex.attr == 'attr'

def test_set():
    @autoprop
    class Example: #
        def set_attr(self, attr): #
            self._attr = 'set ' + attr

    ex = Example()
    ex.attr = 'attr'
    assert ex._attr == 'set attr'

def test_del():
    @autoprop
    class Example: #
        def del_attr(self): #
            self._attr = 'del'

    ex = Example()
    del ex.attr
    assert ex._attr == 'del'

def test_get_set():
    @autoprop
    class Example: #
        def get_attr(self): #
            return self._attr
        def set_attr(self, attr): #
            self._attr = 'set ' + attr

    ex = Example()
    ex.attr = 'attr'
    assert ex.attr == 'set attr'

def test_get_set_del():
    @autoprop
    class Example: #
        def get_attr(self): #
            return self._attr
        def set_attr(self, attr): #
            self._attr = 'set ' + attr
        def del_attr(self): #
            self._attr = 'del'

    ex = Example()
    ex.attr = 'attr'
    assert ex.attr == 'set attr'
    del ex.attr
    assert ex.attr == 'del'

def test_ignore_similar_names():
    @autoprop
    class Example: #
        def getattr(self): #
            return 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

def test_ignore_empty_names():
    @autoprop
    class Example: #
        def get_(self): #
            return 'get'

    ex = Example()
    with pytest.raises(AttributeError):
        getattr(ex, '')

def test_ignore_non_methods():
    @autoprop
    class Example: #
        get_attr = 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

def test_ignore_wrong_args_get():
    @autoprop
    class Example: #
        def get_attr(): #
            return 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

    @autoprop
    class Example: #
        def get_attr(self, more_info): #
            return 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

def test_ignore_wrong_args_set():
    @autoprop
    class Example: #
        def set_attr(self): #
            self._attr = 'no args'

    ex = Example()
    ex.attr = 'attr'
    with pytest.raises(AttributeError):
        assert ex._attr != 'no args'

    @autoprop
    class Example: #
        def set_attr(self, new_value, more_info): #
            self._attr = 'two args'

    ex = Example()
    ex.attr = 'attr'
    with pytest.raises(AttributeError):
        assert ex._attr != 'two args'

def test_ignore_wrong_args_del():
    @autoprop
    class Example: #
        def del_attr(): #
            pass

    ex = Example()
    with pytest.raises(AttributeError):
        del ex.attr

    @autoprop
    class Example: #
        def del_attr(self, more_info): #
            pass

    ex = Example()
    with pytest.raises(AttributeError):
        del ex.attr

def test_ignore_wrong_args_inherited():
    # Tricky case where the parent class has methods with the right names to be 
    # used in autoprops, but not the right arguments.  The child class should 
    # not use these parent class methods.

    @autoprop
    class Parent: #
        def get_attr(self, arg1):
            pass
        def set_attr(self, arg1, arg2): #
            pass

    @autoprop
    class GetChild(Parent): #
        def get_attr(self): #
            return 'attr-child'

    c = GetChild()
    assert c.attr == 'attr-child'
    with pytest.raises(AttributeError):
        c.attr = 'x'

    @autoprop
    class SetChild(Parent): #
        def set_attr(self, value): #
            self._attr = value

    c = SetChild()
    c.attr = 'x'
    assert c._attr == 'x'
    with pytest.raises(AttributeError):
        c.attr

def test_dont_overwrite_existing_attributes():
    @autoprop
    class Example: #
        attr = 'class var'
        def get_attr(self): #
            return 'attr'

    ex = Example()
    assert ex.attr == 'class var'

def test_overwrite_inherited_attributes():
    class Parent: #
        attr = 'parent'

    @autoprop
    class Child(Parent): #
        def get_attr(self): #
            return 'child'

    p = Parent()
    c = Child()

    assert p.attr == 'parent'
    assert c.attr == 'child'

def test_overwrite_inherited_autoprops_1():
    @autoprop
    class Parent: #
        def get_attr(self): #
            return 'parent'
        def get_overloaded_attr(self): #
            return 'parent'

    @autoprop
    class Child(Parent): #
        def get_overloaded_attr(self): #
            return 'child'

    p = Parent()
    c = Child()

    assert p.attr == 'parent'
    assert p.overloaded_attr == 'parent'
    assert c.attr == 'parent'
    assert c.overloaded_attr == 'child'

def test_overwrite_inherited_autoprops_2():

    @autoprop
    class Parent:
        def __init__(self): #
            self._attr = 'attr'
        def get_attr(self): #
            return self._attr
        def set_attr(self, attr): #
            self._attr = attr

    @autoprop
    class Child1(Parent):
        pass

    @autoprop
    class Child2(Parent):
        def get_attr(self): #
            return self._attr + '-get2'

    @autoprop
    class Child3(Parent):
        def get_attr(self): #
            return self._attr + '-get3'
        def set_attr(self, attr): #
            self._attr = attr + '-set3'

    assert 'attr' in Parent.__dict__
    assert 'attr' not in Child1.__dict__
    assert 'attr' in Child2.__dict__
    assert 'attr' in Child3.__dict__

    p = Parent()
    c1 = Child1()
    c2 = Child2()
    c3 = Child3()

    assert p.attr == 'attr'
    assert c1.attr == 'attr'
    assert c2.attr == 'attr-get2'
    assert c3.attr == 'attr-get3'

    c1.attr = 'xyz'
    c2.attr = 'xyz'
    c3.attr = 'xyz'

    assert c1.attr == 'xyz'
    assert c2.attr == 'xyz-get2'
    assert c3.attr == 'xyz-set3-get3'

def test_partially_inherited_autoprops():
    # If the child only implements one of the accessors, the others should be 
    # picked up from the parent class.

    @autoprop
    class Parent: #
        def get_attr(self): #
            return ['parent get', *self._attr]
        def set_attr(self, value): #
            self._attr = ['parent set', value]
        def del_attr(self): #
            self._attr = ['parent del']

    @autoprop
    class GetChild(Parent): #
        def get_attr(self): #
            return ['child get', *self._attr]

    c = GetChild()
    c.attr = 'x'
    assert c.attr == ['child get', 'parent set', 'x']
    del c.attr
    assert c.attr == ['child get', 'parent del']

    @autoprop
    class SetChild(Parent): #
        def set_attr(self, value): #
            self._attr = ['child set', value]

    c = SetChild()
    c.attr = 'x'
    assert c.attr == ['parent get', 'child set', 'x']
    del c.attr
    assert c.attr == ['parent get', 'parent del']

    @autoprop
    class DelChild(Parent): #
        def del_attr(self): #
            self._attr = ['child del']

    c = DelChild()
    c.attr = 'x'
    assert c.attr == ['parent get', 'parent set', 'x']
    del c.attr
    assert c.attr == ['parent get', 'child del']

def test_optional_arguments():
    @autoprop
    class Example: #
        def get_attr(self, pos=None, *args, **kwargs): #
            return ['get', *self._attr]
        def set_attr(self, new_value, pos=None, *args, **kwargs): #
            self._attr = ['set', new_value]
        def del_attr(self, pos=None, *args, **kwargs): #
            self._attr = ['del']

    ex = Example()
    ex.attr = 'x'
    assert ex.attr == ['get', 'set', 'x']
    del ex.attr
    assert ex.attr == ['get', 'del']

def test_keyword_only_arguments():
    # inspect.getargspec() chokes on methods with keyword-only arguments.

    @autoprop
    class Example: #
        def get_attr(self, *, pos=None): #
            return 'attr'
        def set_attr(self, *, new_value): #
            pass

    ex = Example()
    assert ex.attr == 'attr'

    # Even though the setter has one required argument, it's a named argument, 
    # so it shouldn't count and 'attr' should be defined without a setter.
    with pytest.raises(AttributeError):
        ex.attr = 'setter'

@pytest.mark.xfail
def test_classmethod():
    # I wrote this test when trying to address #1, before realizing that this 
    # probably isn't possible.  I'm leaving the test cases because the syntax 
    # is nice in principle, and maybe I'll find a way to implement it some day.

    @autoprop
    class Example: #

        @classmethod
        def get_attr(cls): #
            return ['get', self._attr]

        @classmethod
        def set_attr(cls, value): #
            cls._attr = ['set', value]

        @classmethod
        def del_attr(cls): #
            cls._attr = ['del']

    Example.attr = 'x'
    assert Example.attr == ['get', 'set', 'x']
    del Example.attr
    assert Example.attr == ['get', 'del']

@pytest.mark.xfail
def test_staticmethod():
    # I wrote this test when trying to address #1, before realizing that this 
    # probably isn't possible.  I'm leaving the test cases because the syntax 
    # is nice in principle, and maybe I'll find a way to implement it some day.

    @autoprop
    class Example: #

        @staticmethod
        def get_attr(): #
            return ['get', self._attr]

        @staticmethod
        def set_attr(value): #
            Example._attr = ['set', value]

        @staticmethod
        def del_attr(): #
            Example._attr = ['del']

    Example.attr = 'x'
    assert Example.attr == ['get', 'set', 'x']
    del Example.attr
    assert Example.attr == ['get', 'del']


@pytest.mark.parametrize('decorator', [autoprop, autoprop.cache])
def test_docstrings(decorator):
    # This is a bit of a tricky case because getting the docstring requires 
    # accessing the `CachedProperty` descriptor as a class attribute, which 
    # changes how the arguments to `__get__()` need to be interpreted.

    @decorator
    class Example: #
        def get_attr(self): #
            "get attr"
            pass
        def set_attr(self, attr): #
            "set attr"
            pass
        def del_attr(self): #
            "del attr"
            pass

    ex = Example()

    # The docstrings on the setter and deleter are ignored, as per the default 
    # behavior of the `property()` decorator.
    assert Example.attr.__doc__ == "get attr"





#!/usr/bin/env python

"""Infer properties from accessor methods."""

import sys
import inspect
import re
import functools
import signature_dispatch
from functools import partial
from collections import defaultdict

__version__ = '3.0.0'

_EXPECTED_NUM_ARGS = {'get': 0, 'set': 1, 'del': 0}
_CACHE_POLICIES = ['dynamic', 'class', 'object', 'property', 'immutable']
_CACHE_ATTR = '_autoprop_cache'
_CACHE_VERSION_ATTR = '_autoprop_cache_version'
_CACHE_POLICY_ATTR = '_autoprop_cache_policy'
_REFRESH_ATTR = '_autoprop_refresh'
_INITIAL_VERSION = 1
_DEFAULT_POLICY = 'property'

class Cache:

    def __init__(self, obj):
        self.obj = obj
        self.cls = obj.__class__

        self.values = {}
        self.cls_versions = {}
        self.obj_versions = {}
        self.attr_touched = {}

    def bind(self, attr):
        return BoundCache(self, attr)

class BoundCache:

    def __init__(self, cache, attr):
        self.cache = cache
        self.attr = attr

    def is_stale(self, policy):
        if policy == 'dynamic':
            return True

        cache = self.cache
        attr = self.attr

        if attr not in cache.values:
            return True

        if policy == 'immutable':
            return False

        if cache.attr_touched.get(attr, False):
            return True

        if policy == 'property':
            return False

        v0 = _INITIAL_VERSION - 1

        if cache.obj_versions.get(attr, v0) < get_cache_version(cache.obj):
            return True

        if policy == 'object':
            return False

        assert policy == 'class'

        for base in cache.cls.__mro__:
            my_version = cache.cls_versions.get(base, {}).get(attr, v0)
            base_version = get_cache_version(base)
            if my_version < base_version:
                return True

        return False

    def get(self):
        return self.cache.values[self.attr]

    def update(self, value):
        cache = self.cache
        attr = self.attr

        cache.values[attr] = value
        cache.attr_touched[attr] = False
        cache.obj_versions[attr] = get_cache_version(cache.obj)

        # Build this dictionary on the fly to be robust against the (remote) 
        # possibility that `__mro__` could change at runtime, e.g.:
        # https://stackoverflow.com/questions/20822850/change-python-mro-at-runtime 

        for base in cache.cls.__mro__:
            cache.cls_versions.setdefault(base, {})[attr] = \
                    get_cache_version(base)

    def touch(self):
        self.cache.attr_touched[self.attr] = True

class CachedProperty(property):

    def __init__(self, name, getter, setter, deleter, policy):
        super().__init__(getter, setter, deleter)
        self.name = name
        self.policy = policy

    def __get__(self, obj, owner=None):
        # Class attribute access (e.g. for docstrings): 
        if obj is None:
            return super().__get__(obj, owner)

        # Instance attribute access:
        cache = self.get_bound_cache(obj)

        if cache.is_stale(self.policy):
            value = super().__get__(obj, owner)
            cache.update(value)

        return cache.get()

    def __set__(self, obj, value):
        cache = self.get_bound_cache(obj)
        cache.touch()
        return super().__set__(obj, value)

    def __delete__(self, obj):
        cache = self.get_bound_cache(obj)
        cache.touch()
        return super().__delete__(obj)

    def get_bound_cache(self, obj):
        cache = get_cache(obj)
        return cache.bind(self.name)

    def getter_replacement(self):
        def getter(obj, *args, **kwargs):
            if not args and not kwargs:
                return self.__get__(obj, None)
            else:
                return self.fget(obj, *args, **kwargs)
        return getter

def autoprop(cls):
    return _make_autoprops(cls)

def cache(*args, **kwargs):

    def wrapper(x, policy):
        if inspect.isclass(x):
            return _make_autoprops(x, cache=True, default_policy=policy)
        else:
            return _assign_policy(x, policy=policy)

    dispatch = signature_dispatch()

    @dispatch
    def decorator(*, policy):
        _check_policy(policy)
        return partial(wrapper, policy=policy)

    @dispatch
    def decorator():
        return partial(wrapper, policy=_DEFAULT_POLICY)

    @dispatch
    def decorator(f):
        return wrapper(f, _DEFAULT_POLICY)

    return decorator(*args, **kwargs)

def dynamic(f):
    return cache(policy='dynamic')(f)

def immutable(f):
    return cache(policy='immutable')(f)

def policy(policy):
    _check_policy(policy)

    def wrapper(f):
        _assign_policy(f, policy)
        return f

    return wrapper

def refresh(f):
    setattr(f, _REFRESH_ATTR, True)
    return f

def get_cache(obj):
    try:
        cache = getattr(obj, _CACHE_ATTR)
    except AttributeError:
        cache = Cache(obj)
        setattr(obj, _CACHE_ATTR, cache)

    return cache

def get_cache_version(obj):
    return getattr(obj, _CACHE_VERSION_ATTR, _INITIAL_VERSION)

def clear_cache(obj):
    try:
        delattr(obj, _CACHE_ATTR)
    except AttributeError:
        pass


def _make_autoprops(cls, *, cache=False, default_policy=_DEFAULT_POLICY):
    accessors = defaultdict(dict)

    def get_accessor(name, prefix):
        # If we found the accessor in this class, return it.
        try:
            return accessors[name][prefix]

        # Otherwise, look for a suitable method in parent classes.
        except KeyError:

            try:
                full_name = '%s_%s' % (prefix, name)
                attr = getattr(cls, full_name)
            except AttributeError:
                return None
            else:
                return attr if _is_accessor(full_name, attr) else None

    def increment_cache_version(f, predicate=lambda *args, **kwargs: True):
        if f is None:
            return f

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if predicate(*args, **kwargs):
                if args and isinstance(args[0], cls):
                    obj = args[0]
                else:
                    obj = cls

                version = get_cache_version(obj)
                setattr(obj, _CACHE_VERSION_ATTR, version + 1)

            return f(*args, **kwargs)

        return wrapper

    for name, attr in cls.__dict__.items():
        # Because we are iterating through `__dict__`, class/static methods 
        # will not be bound and will not appear as functions, which is good 
        # because we cannot make class properties without metaclasses.

        key = _is_accessor(name, attr)
        if key:
            name, prefix = key
            accessors[name][prefix] = attr

            if hasattr(attr, _CACHE_POLICY_ATTR):
                if not cache:
                    raise ValueError("\n".join([
                        f"cache disabled: can't specify cache policies",
                        f"class: {cls.__qualname__}",
                        f"method: {attr.__qualname__}",
                        f"decorate the class with `@autoprop.cache` or `@autoprop.dynamic` to enable the cache.",
                    ]))


        if getattr(attr, _REFRESH_ATTR, False):
            if not cache:
                raise ValueError("\n".join([
                    f"cache disabled: can't use `@autoprop.refresh`",
                    f"class: {cls.__qualname__}",
                    f"method: {attr.__qualname__}",
                    f"decorate the class with `@autoprop.cache` or `@autoprop.dynamic` to enable the cache.",
                ]))
            setattr(cls, name, increment_cache_version(attr))

        # We have to dig inside static/class methods to see if the underlying 
        # function is supposed to get decorated.

        if isinstance(attr, (classmethod, staticmethod)):
            if getattr(attr.__func__, _REFRESH_ATTR, False):
                if not cache:
                    raise ValueError("\n".join([
                        f"cache disabled: can't use `@autoprop.refresh`",
                        f"class: {cls.__qualname__}",
                        f"method: {attr.__func__.__qualname__}",
                        f"decorate the class with `@autoprop.cache` or `@autoprop.dynamic` to enable the cache.",
                    ]))
                attr = type(attr)(increment_cache_version(attr.__func__))
                setattr(cls, name, attr)

    for name in accessors:

        # Don't overwrite any attributes defined in this class.  Attributes 
        # defined in superclasses may be shadowed.

        if name in cls.__dict__:
            continue

        getter  = get_accessor(name, 'get')
        setter  = get_accessor(name, 'set')
        deleter = get_accessor(name, 'del')

        if not cache:
            # Mark the getter as 'dynamic'.  This won't have any affect on this 
            # class, but if a subclass with caching enabled incorporates this 
            # getter into an autoprop, this policy will ensure that the getter 
            # remains un-cached (as its author presumably intended).
            if getter:
                dynamic(getter)

            prop = property(getter, setter, deleter)

        else:
            setter = increment_cache_version(setter)
            deleter = increment_cache_version(deleter)
            policy = getattr(getter, _CACHE_POLICY_ATTR, default_policy)

            if policy == 'immutable':
                if getter and setter:
                    raise ValueError("\n".join([
                        f"can't specify setter for immutable property",
                        f"property: {cls.__qualname__}.{name}",
                        f"immutable getter: {getter.__qualname__}",
                        f"setter: {setter.__qualname__}",
                    ]))
                if getter and deleter:
                    raise ValueError("\n".join([
                        f"can't specify deleter for immutable property",
                        f"property: {cls.__qualname__}.{name}",
                        f"immutable getter: {getter.__qualname__}",
                        f"deleter: {deleter.__qualname__}",
                    ]))

            prop = CachedProperty(name, getter, setter, deleter, policy)

            if getter:
                setattr(cls, f'get_{name}', prop.getter_replacement())

        setattr(cls, name, prop)

    if cache:

        # Automatically increment the cache on attribute assignment.  
        # 
        # Don't do the same for attribute deletion, because (i) classes are not 
        # supposed to implement `__delattr__()` unless they actually support it 
        # and (ii) it's easy to get this behavior by decorating `__delattr__()` 
        # with `@autoprop.refresh`.  Decorating `__setattr__()` is not so 
        # easy because steps must be taken to avoid infinite recursion.

        try:
            __setattr__ = cls.__dict__['__setattr__']
        except KeyError:
            def __setattr__(self, name, value):
                super(cls, self).__setattr__(name, value)

        def predicate(self, name, value):
            return not name.startswith('_autoprop')

        __setattr__ = increment_cache_version(__setattr__, predicate)
        setattr(cls, '__setattr__', __setattr__)

    return cls

def _check_policy(policy):
    if policy not in _CACHE_POLICIES:
        expected = ', '.join(repr(p) for p in _CACHE_POLICIES)
        raise ValueError(f"unknown policy {policy!r}, expected one of: {expected}")

def _assign_policy(f, policy):
    if not f.__name__.startswith('get_'):
        raise ValueError(f"can't cache {f.__qualname__}; it's not a getter")

    setattr(f, _CACHE_POLICY_ATTR, policy)
    return f

def _is_accessor(name, attr):
    if not inspect.isfunction(attr):
        return False

    accessor_match = re.match('(get|set|del)_(.+)', name)
    if not accessor_match:
        return False

    def is_required(p):
        positional = p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD
        return p.kind in positional and p.default is p.empty

    # The attributes passed to this function must be accessed via `getattr()` 
    # (e.g. as opposed to `cls.__dict__[]`) because static/class methods are 
    # not callable until bound to the class.

    prefix, name = accessor_match.groups()
    sig = inspect.signature(attr)
    params = sig.parameters.values()
    num_args = sum(map(is_required, params)) - 1

    if num_args != _EXPECTED_NUM_ARGS[prefix]:
        return False

    return name, prefix

# Hack to make the module directly usable as a decorator.  Only works for 
# python 3.5 or higher.  See this Stack Overflow post:
# https://stackoverflow.com/questions/1060796/callable-modules

class CallableModule(sys.modules[__name__].__class__):

    def __call__(self, *args, **kwargs):
        return autoprop(*args, **kwargs)

sys.modules[__name__].__class__ = CallableModule




#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

install_requires = \
['signature_dispatch']

extras_require = \
{'doc': ['sphinx', 'sphinx_rtd_theme', 'autoclasstoc'],
 'test': ['pytest', 'pytest-cov', 'coveralls']}

setup(name='autoprop',
      version='3.0.0',
      description='Infer properties from accessor methods.',
      author='Kale Kundert',
      author_email='kale@thekunderts.net',
      url='https://github.com/kalekundert/autoprop',
      py_modules=['autoprop'],
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='~=3.6',
     )
